
// NoMoSmooth
// Avisynth	filter for motion-adaptive spatio-temporal smoothing
//
// By Ross Thomas <ross@grinfinity.com>
//
// There is	no copyright on	this code, and there are no	conditions
// on its distribution or use. Do with it what you will.

// YUY2	version	for	2.0	and	2.5

class NoMoSmooth_YUY2 : public GenericVideoFilter
{
public:
	NoMoSmooth_YUY2(PClip _child, int motion_threshold, int temporal_radius,
		int temporal_threshold, int spatial_radius, int spatial_threshold,
		bool show, IScriptEnvironment* env);
	virtual ~NoMoSmooth_YUY2();
	PVideoFrame __stdcall GetFrame(int n, IScriptEnvironment* env);

private:
	const int motion_threshold;
	const int temporal_radius, temporal_threshold;
	const int spatial_radius, spatial_threshold;
	const bool show;

	const unsigned char **p;
	int temporal_diameter, spatial_diameter, sr2;

	// Lookups

	BYTE** spatial_sum_lookup;
	BYTE** spatial_div_lookup;
	BYTE** temporal_sum_lookup;
	BYTE** temporal_div_lookup;
	int** div_lookup;
};

NoMoSmooth_YUY2::NoMoSmooth_YUY2(PClip _child, int _motion_threshold,
								 int _temporal_radius, int _temporal_threshold,
								 int _spatial_radius, int _spatial_threshold,
								 bool _show, IScriptEnvironment* env)
	: GenericVideoFilter(_child), motion_threshold(_motion_threshold),
	  temporal_radius(_temporal_radius), temporal_threshold(_temporal_threshold),
	  spatial_radius(_spatial_radius), spatial_threshold(_spatial_threshold),
	  show(_show)
{
	assert(_child);
	assert(motion_threshold > 0);
	assert(temporal_radius > 0);
	assert(temporal_threshold > 0);
	assert(spatial_radius > 0);
	assert(spatial_threshold > 0);

	temporal_diameter = temporal_radius * 2 + 1;
	spatial_diameter = spatial_radius * 2 + 1;
	sr2 = spatial_radius * 2;
	p = new const unsigned char*[temporal_diameter];
	if(!p)
		env->ThrowError(MEMORY_ERROR);

	// Smoothing lookups: Amount to add to the sum and div
	// where the pixel is within the threshold, or zero

	spatial_sum_lookup = new BYTE*[256];
	spatial_div_lookup = new BYTE*[256];
	temporal_sum_lookup = new BYTE*[256];
	temporal_div_lookup = new BYTE*[256];
	if(!spatial_sum_lookup || !spatial_div_lookup ||
		!temporal_sum_lookup || !temporal_div_lookup)
		env->ThrowError(MEMORY_ERROR);
	for(int i = 0; i < 256; ++i)
	{
		spatial_sum_lookup[i] = new BYTE[256];
		spatial_div_lookup[i] = new BYTE[256];
		temporal_sum_lookup[i] = new BYTE[256];
		temporal_div_lookup[i] = new BYTE[256];
		if(!spatial_sum_lookup[i] || !spatial_div_lookup[i] ||
			!temporal_sum_lookup[i] || !temporal_div_lookup[i])
			env->ThrowError(MEMORY_ERROR);
		for(int j = 0; j < 256; ++j)
		{
			if(abs(i - j) <= spatial_threshold)
			{
				spatial_sum_lookup[i][j] = j;
				spatial_div_lookup[i][j] = 1;
			} else
			{
				spatial_sum_lookup[i][j] = 0;
				spatial_div_lookup[i][j] = 0;
			}
			if(abs(i - j) <= temporal_threshold)
			{
				temporal_sum_lookup[i][j] = j;
				temporal_div_lookup[i][j] = 1;
			} else
			{
				temporal_sum_lookup[i][j] = 0;
				temporal_div_lookup[i][j] = 0;
			}
		}
	}
	assert(256 == i);

	// Divisor lookup: For averaging, where div_lookup[a][b] is
	// equivalent to a / b

	int maxnum, maxden;
	if(temporal_diameter > spatial_diameter * spatial_diameter)
	{
		maxnum = 256 * temporal_diameter;
		maxden = temporal_diameter + 1;
	} else
	{
		maxnum = 256 * spatial_diameter * spatial_diameter;
		maxden = spatial_diameter * spatial_diameter + 1;
	}
	div_lookup = new int*[maxnum];
	if(!div_lookup)
		env->ThrowError(MEMORY_ERROR);
	for(int i2 = 0; i2 < maxnum; ++i2)
	{
		div_lookup[i2] = new int[maxden + 1];
		if(!div_lookup[i2])
			env->ThrowError(MEMORY_ERROR);
		for(int j = 1; j <= maxden; ++j)
			div_lookup[i2][j] = i2 / j;
		assert(j == maxden + 1);
	}
	assert(i2 == maxnum);
}

NoMoSmooth_YUY2::~NoMoSmooth_YUY2()
{
	assert(p);
	delete [] p;

	assert(spatial_sum_lookup);
	assert(spatial_div_lookup);
	assert(temporal_sum_lookup);
	assert(temporal_div_lookup);
	for(int i = 0; i < 256; ++i)
	{
		assert(spatial_sum_lookup[i]);
		delete [] spatial_sum_lookup[i];
		assert(spatial_div_lookup[i]);
		delete [] spatial_div_lookup[i];
		assert(temporal_sum_lookup[i]);
		delete [] temporal_sum_lookup[i];
		assert(temporal_div_lookup[i]);
		delete [] temporal_div_lookup[i];
	}
	delete [] spatial_sum_lookup;
	delete [] spatial_div_lookup;
	delete [] temporal_sum_lookup;
	delete [] temporal_div_lookup;

	assert(div_lookup);
	int max;
	if(temporal_diameter > spatial_diameter * spatial_diameter)
		max = 256 * temporal_diameter;
	else
		max = 256 * spatial_diameter * spatial_diameter;
	for(int i3 = 0; i3 < max; ++i3)
	{
		assert(div_lookup[i3]);
		delete [] div_lookup[i3];
	}
	assert(i3 == max);
	delete [] div_lookup;
}

PVideoFrame NoMoSmooth_YUY2::GetFrame(int n, IScriptEnvironment* env)
{
	assert(n >= 0);
	assert(n < vi.num_frames);
	assert(env);

	// Get current frame

	PVideoFrame fcurr = child->GetFrame(n, env);
	assert(fcurr);

	// Since we need to temporally smooth with temporal_radius
	// frames in each direction, we need to skip the first
	// and last frames within that range

	if(n < temporal_radius || n > vi.num_frames - temporal_radius)
		return fcurr;

	// We prefetch all frames required to temporally smooth,
	// which means if e.g. n=10 and temporal_radius=2 we have
	// a frame_cnt=5 buffer:
	//
	//  Index: 0    1    2    3    4
	// Offset: n-2  n-1  n    n+1  n+2
	//  Frame: 8    9    10   11   12

	for(int i = 0; i < temporal_diameter; ++i)
	{
		assert(n - temporal_radius + i >= 0);
		assert(n - temporal_radius + i < vi.num_frames);
		PVideoFrame ftemp = child->GetFrame(n - temporal_radius + i, env);
		assert(ftemp);
		p[i] = ftemp->GetReadPtr();
		assert(p[i]);
	}
	assert(i == temporal_diameter);

	// We make aliases to the current, previous and next frames
	// for improved comprehension

	const unsigned char *&pcurr = p[temporal_radius],
		*&pprev = p[temporal_radius - 1], *&pnext = p[temporal_radius + 1];

	// Create new destination frame

	PVideoFrame fdest = env->NewVideoFrame(vi);
	assert(fdest);
	unsigned char* pdest = fdest->GetWritePtr();
	assert(pdest);

	// Get constants. Width must be mod 2, and since row_size is
	// twice width in YUY2, row_size must be mod 4

	const int row_size = fcurr->GetRowSize(), height = fcurr->GetHeight(),
		pitch = fcurr->GetPitch(), dpitch = fdest->GetPitch();

	assert(row_size > 0);
	assert(0 == row_size % 4);
	assert(height > 0);
	assert(pitch > 0);
	assert(dpitch > 0);

	// Algorithm:
	//
	// Detect motion between current and previous frames.
	// We do this by checking each pixel in a 3x3 matrix
	// against its corresponding pixel in the previous
	// frame. We sum all the differences then take the
	// absolute value (as opposed to summing the absolute
	// value of the differences). This gives us an
	// approximation of how many pixels are moving in the
	// same direction.
	//
	// Pixels that are considered in motion will be
	// spatially smoothed. Those that aren't are smoothed
	// temporally.
	//
	// We do not process the very edges since we do not
	// want to stray outside the frame while checking our
	// 3x3 matrix, so copy all the edges up front:

	// Top

	env->BitBlt(pdest, dpitch, pcurr, pitch, row_size, spatial_radius);

	// Bottom

	env->BitBlt(pdest + (height - spatial_radius) * dpitch, dpitch,
		pcurr + (height - spatial_radius) * pitch, pitch, row_size, spatial_radius);

	// Left

	env->BitBlt(pdest, dpitch, pcurr, pitch, sr2, height);

	// Right

	env->BitBlt(pdest + row_size - sr2, dpitch, pcurr + row_size - sr2,
		pitch, sr2, height);

	// Advance pointers

	for(int i4 = 0; i4 < temporal_diameter; ++i4)
		p[i4] += pitch * spatial_radius;
	pdest += dpitch * spatial_radius;

	for(int y = height - sr2; y > 0; --y)
	{
		int xmax = row_size - sr2;
		for(int x = sr2; x < xmax; x += 2)
		{
			// Copy chroma

			pdest[x + 1] = pcurr[x + 1];

			// Detect pixels suitable for smoothing (defined
			// as pixel fluctuating over three frames in such
			// a way as to be either greater than or less than
			// both its surrounding two pixels)

			int prevdiff = pprev[x] - pcurr[x], nextdiff = pnext[x] - pcurr[x];
			if((prevdiff < 0 && nextdiff > 0) || (prevdiff > 0 && nextdiff < 0))
			{
				pdest[x] = pcurr[x];
				continue;
			}

			// Handy pointers to top, centre and bottom
			// rows of 3x3 matrix

			const unsigned char* pprevt = pprev - pitch + x,
				*pprevc = pprev + x,
				*pprevb = pprev + pitch + x,
				*pcurrt = pcurr - pitch + x,
				*pcurrc = pcurr + x,
                *pcurrb = pcurr + pitch + x;

			// Detect motion

			int mo = abs(
				(pprevt[-2] - pcurrt[-2]) +
				(pprevt[0] - pcurrt[0]) +
				(pprevt[2] - pcurrt[2]) +
				(pprevc[-2] - pcurrc[-2]) +
				(pprevc[0] - pcurrc[0]) +
				(pprevc[2] - pcurrc[2]) +
				(pprevb[-2] - pcurrb[-2]) +
				(pprevb[0] - pcurrb[0]) +
				(pprevb[2] - pcurrb[2])
			);

			if(mo > motion_threshold)
			{
				if(show)
				{
					// Highlight in show mode

					pdest[x] = 150;
					continue;
				}

				// Spatially smooth

				int l = pcurrc[0], sum = 0, div = 0;
				const unsigned char* ppmax = pcurr + x + spatial_radius * pitch;
				for(const unsigned char* pp = pcurr + x - spatial_radius * pitch;
					pp <= ppmax; pp += pitch)
				{
					for(int j = -sr2; j <= sr2; j += 2)
					{
						int ll = pp[j];
						sum += spatial_sum_lookup[l][ll];
						div += spatial_div_lookup[l][ll];
					}
					assert(j == sr2 + 2);
				}
				assert(pp == ppmax + pitch);

				// Average

				assert(div > 0);
				assert(div <= spatial_diameter * spatial_diameter);
				assert(sum >= 0);
				assert(sum < 256 * spatial_diameter * spatial_diameter);

				pdest[x] = div_lookup[sum][div];
			} else
			{
				if(show)
				{
					// Straight copy in show mode

					pdest[x] = pcurrc[0];
					continue;
				}

				// Temporally smooth

				int sum = 0, div = 0, l = pcurrc[0];
				for(int i = 0; i < temporal_diameter; ++i)
				{
					int ll = p[i][x];
					sum += temporal_sum_lookup[l][ll];
					div += temporal_div_lookup[l][ll];
				}
				assert(i == temporal_diameter);

				// Average

				assert(div > 0);
				assert(div <= temporal_diameter);
				assert(sum >= 0);
				assert(sum < 256 * temporal_diameter);

				pdest[x] = div_lookup[sum][div];
			}
		}
		assert(x == xmax);

		// Advance

		for(int i = 0; i < temporal_diameter; ++i)
			p[i] += pitch;
		assert(i == temporal_diameter);
		pdest += dpitch;
	}
	assert(y == 0);

	// Done

	assert(fdest);
	return fdest;
}
